@@ -1,4 +1,4 @@
-# Copyright 2024 IQM
+# Copyright 2024-2025 IQM
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -80,6 +80,8 @@ from iqm.cpc.compiler.dd import STANDARD_DD_STRATEGY, insert_dd_sequences, merge
 from iqm.cpc.compiler.errors import (
     CalibrationError,
     CircuitError,
+    ClientError,
+    UnknownCircuitExecutionOptionError,
     UnknownHardwareComponentError,
     UnknownLogicalQubitError,
 )
@@ -230,7 +232,7 @@ def _fix_implementation_and_locus(
         inst.locus = locus
         circuit_components.update(locus)
         if op.arity == 2:
-            circuit_component_pairs.add(locus)
+            circuit_component_pairs.add(locus)  # type: ignore[arg-type]
         circuit_gate_loci.setdefault(inst.name, {}).setdefault(inst.implementation, Counter()).update([locus])
 
     return CircuitMetrics(
@@ -246,6 +248,7 @@ def _build_readout_mappings(
     builder: ScheduleBuilder,
     measurement_mode: MeasurementMode,
     heralding_mode: HeraldingMode,
+    convert_terminal_measurements: bool,
 ) -> tuple[tuple[ReadoutMapping, ...], tuple[tuple[str, ...], ...]]:
     """Create a readout mapping for each circuit, add required measurement instructions.
 
@@ -262,6 +265,8 @@ def _build_readout_mappings(
         builder: schedule builder object, encapsulating station properties and gate calibration data
         measurement_mode: measurement mode to be used
         heralding_mode: type of heralding used
+        convert_terminal_measurements: whether to convert terminal measurements into the ``"measure_fidelity"``
+            operation.
 
     Returns:
         readout mappings, heralded components
@@ -272,7 +277,10 @@ def _build_readout_mappings(
 
     # find out which components have measurement data
     components_that_can_be_measured = frozenset(
-        q for impl_loci in builder.calibration["measure"].values() for locus in impl_loci for q in locus
+        q
+        for impl_loci in builder.calibration["measure"].values()
+        for locus in impl_loci
+        for q in locus  # type: ignore[union-attr]
     )
 
     # Figure out which qubits should always be measured on the QPU in the final measurement.
@@ -283,7 +291,7 @@ def _build_readout_mappings(
         # no extras on top of the measurements actually in each circuit
         final_components_to_measure = frozenset()
     else:
-        raise ValueError(f"Unknown measurement mode: {measurement_mode}")
+        raise UnknownCircuitExecutionOptionError(f"Unknown measurement mode: {measurement_mode}")
 
     for idx, c in enumerate(circuits):
         try:
@@ -318,13 +326,21 @@ def _build_readout_mappings(
                         final_measurements.append(inst)
                 eclipsed_qubits.update(inst.locus)
 
+            if convert_terminal_measurements:
+                final_measurements_name = "measure_fidelity"
+                for measurement in final_measurements:
+                    measurement.name = final_measurements_name
+            else:
+                final_measurements_name = "measure"
             # Add a single final measurement instruction for qubits that still need to be measured
             # (order does not matter). this measurement instruction is not reflected in circuit_metrics,
             # nor are the results returned by SC. NOTE we assume that iqm-pulse is clever enough to
             # combine the separate final measurement instructions into simultaneous ReadoutTriggers.
             if missing := final_components_to_measure - set(q for inst in final_measurements for q in inst.locus):
                 # TODO which implementation to use? Now we leave it to None.
-                c.instructions += (CircuitOperation("measure", tuple(missing), {"key": MEASUREMENT_MODE_KEY}),)
+                c.instructions += (
+                    CircuitOperation(final_measurements_name, tuple(missing), {"key": MEASUREMENT_MODE_KEY}),
+                )
 
             if heralding_mode != HeraldingMode.NONE:
                 # heralding must return results for the components used in the circuit, if they can be measured
@@ -359,7 +375,7 @@ def _get_op_calibration_errors(calibration: OpCalibrationDataTree, ops: QuantumO
 
     """
 
-    def diff_dicts(cal_data: dict[str, Any], impl_parameters: dict[str, Any], path) -> None | str:
+    def diff_dicts(cal_data: dict[str, Any], impl_parameters: dict[str, Any], path) -> None | str:  # noqa: ANN001
         """Compare the calibration"""
         full = set(impl_parameters)
         have = set(cal_data)
@@ -409,22 +425,22 @@ def _get_op_calibration_errors(calibration: OpCalibrationDataTree, ops: QuantumO
                 # since OILCalibrationData can have nested dicts, we do a recursive diff
                 error = diff_dicts(merge_dicts(default_cal_data, cal_data), impl.parameters, [])
                 if error is not None:
-                    errors[(op_name, impl_name, locus)] = error
+                    errors[(op_name, impl_name, locus)] = error  # type: ignore[index]
                     continue
 
                 n_components = len(locus)
                 arity = op.arity
                 if arity == 0:
                     if n_components != 1:
-                        errors[(op_name, impl_name, locus)] = (
+                        errors[(op_name, impl_name, locus)] = (  # type: ignore[index]
                             f"{op_name}.{impl_name} at {locus}: for zero-arity operations, "
                             "calibration data must be provided for single-component loci only"
                         )
                 elif n_components != arity:
-                    errors[(op_name, impl_name, locus)] = (
+                    errors[(op_name, impl_name, locus)] = (  # type: ignore[index]
                         f"{op_name}.{impl_name} at {locus}: locus must have {arity} component(s)"
                     )
-    return errors
+    return errors  # type: ignore[return-value]
 
 
 def merge_multiplexed_timeboxes(circuit_box: TimeBox) -> TimeBox:
@@ -467,7 +483,7 @@ def merge_multiplexed_timeboxes(circuit_box: TimeBox) -> TimeBox:
             if pending:
                 if disjoint_boxes(pending, gate_box):
                     # Pending box and new candidate have disjoint loci, merge is possible.
-                    pending = pending + gate_box.children[0]
+                    pending = pending + gate_box.children[0]  # type: ignore[assignment]
                     continue
                 # Pending box collides with the new candidate, so we must place it immediately and continue with
                 # the new candidate.
@@ -478,7 +494,7 @@ def merge_multiplexed_timeboxes(circuit_box: TimeBox) -> TimeBox:
             if pending and not disjoint_boxes(pending, gate_box):
                 placed_boxes.append(pending)
                 pending = None
-            placed_boxes.append(gate_box)
+            placed_boxes.append(gate_box)  # type: ignore[arg-type]
 
     if pending:
         placed_boxes.append(pending)
@@ -493,7 +509,7 @@ def merge_multiplexed_timeboxes(circuit_box: TimeBox) -> TimeBox:
 # Passes of the standard stages
 # CIRCUIT-LEVEL PASSES
 @compiler_pass
-def validate_execution_options(circuits: Iterable[Circuit_], options: CircuitExecutionOptions):
+def validate_execution_options(circuits: Iterable[Circuit_], options: CircuitExecutionOptions):  # noqa: ANN201
     """Validate the circuit execution options (only some combinations make sense)."""
     if options.move_gate_frame_tracking == MoveGateFrameTrackingMode.FULL and options.move_gate_validation not in [
         MoveGateValidationMode.STRICT,
@@ -505,7 +521,7 @@ def validate_execution_options(circuits: Iterable[Circuit_], options: CircuitExe
 
 
 @compiler_pass
-def map_old_operations(circuits: Iterable[Circuit_]):
+def map_old_operations(circuits: Iterable[Circuit_]):  # noqa: ANN201
     """Map backwards-compatible aliases for quantum operation names into the current name."""
     for c in circuits:
         _map_old_operation_names(c.instructions)
@@ -514,7 +530,7 @@ def map_old_operations(circuits: Iterable[Circuit_]):
 
 
 @compiler_pass
-def validate_circuits(circuits: Iterable[Circuit_], builder: ScheduleBuilder):
+def validate_circuits(circuits: Iterable[Circuit_], builder: ScheduleBuilder):  # noqa: ANN201
     """Validate the contents of the quantum circuits."""
     for idx, c in enumerate(circuits):
         try:
@@ -525,7 +541,7 @@ def validate_circuits(circuits: Iterable[Circuit_], builder: ScheduleBuilder):
 
 
 @compiler_pass
-def map_components(
+def map_components(  # noqa: ANN201
     circuits: Iterable[Circuit_],
     builder: ScheduleBuilder,
     component_mapping: dict[str, str],
@@ -538,7 +554,7 @@ def map_components(
             _map_components_in_instructions(
                 component_mapping,
                 c.instructions,
-                device_components=device_components,
+                device_components=device_components,  # type: ignore[arg-type]
             )
         except CircuitError as exc:
             raise CircuitError(f"Circuit {idx}: {exc}") from exc
@@ -586,7 +602,12 @@ def derive_readout_mappings(
 ) -> tuple[list[Circuit_], dict[str, Any]]:
     """Derive mapping between station acquisition labels and user's measurement keys."""
     readout_mappings, heralded_components = _build_readout_mappings(
-        circuits, tuple(circuit_metrics), builder, options.measurement_mode, options.heralding_mode
+        circuits,
+        tuple(circuit_metrics),
+        builder,
+        options.measurement_mode,
+        options.heralding_mode,
+        options.convert_terminal_measurements,
     )
     return list(circuits), {"readout_mappings": readout_mappings, "heralded_components": heralded_components}
 
@@ -595,12 +616,16 @@ def derive_readout_mappings(
 @compiler_pass
 def resolve_circuits(circuits: Iterable[Circuit_], builder: ScheduleBuilder) -> list[TimeBox]:
     """Resolve the circuits to timeboxes."""
-    return [builder.circuit_to_timebox(c.instructions, name=c.name) for c in circuits]
+    try:
+        timeboxes = [builder.circuit_to_timebox(c.instructions, name=c.name) for c in circuits]
+    except ValueError as exc:
+        raise ClientError(f"{exc}") from exc
+    return timeboxes
 
 
 # TIMEBOX-LEVEL PASSES
 @compiler_pass
-def multiplex_readout(timeboxes: Iterable[TimeBox]):
+def multiplex_readout(timeboxes: Iterable[TimeBox]):  # noqa: ANN201
     """Merge any MultiplexedProbeTimeBoxes inside a TimeBox representing a circuit."""
     return [merge_multiplexed_timeboxes(circuit_box) for circuit_box in timeboxes]
 
@@ -622,12 +647,16 @@ def prepend_heralding(
     """Add the heralding measurement timebox to all circuits."""
     if options.heralding_mode != HeraldingMode.ZEROS:
         return list(copy(timeboxes))
-    return [
-        builder.get_implementation("measure", heralded_components[circuit_idx])(key=HERALDING_KEY)
-        + builder.wait(heralded_components[circuit_idx], BUFFER_AFTER_MEASUREMENT_PROBE)
-        | box
-        for circuit_idx, box in enumerate(timeboxes)
-    ]
+    try:
+        timeboxes = [
+            builder.get_implementation("measure", heralded_components[circuit_idx])(key=HERALDING_KEY)
+            + builder.wait(heralded_components[circuit_idx], BUFFER_AFTER_MEASUREMENT_PROBE)
+            | box
+            for circuit_idx, box in enumerate(timeboxes)
+        ]
+    except ValueError as exc:
+        raise ClientError(f"{exc}") from exc
+    return timeboxes
 
 
 @compiler_pass
@@ -640,16 +669,19 @@ def prepend_reset(
     """Add a reset timebox to all circuits."""
     if "reset_wait" not in builder.calibration:  # backwards compatibility
         return timeboxes
-    new_boxes: list[TimeBox] = []
-    for box, metrics in zip(timeboxes, circuit_metrics):
-        reset_components = tuple(metrics.components)
-        if options.active_reset_cycles is None:
-            reset_box = TimeBox.composite([builder.get_implementation("reset_wait", reset_components)()])
-        else:
-            reset_box = TimeBox.composite(
-                [builder.get_implementation("reset", reset_components)()] * options.active_reset_cycles
-            )
-        new_boxes.append(reset_box + box)
+    try:
+        new_boxes: list[TimeBox] = []
+        for box, metrics in zip(timeboxes, circuit_metrics):
+            reset_components = tuple(metrics.components)
+            if options.active_reset_cycles is None:
+                reset_box = TimeBox.composite([builder.get_implementation("reset_wait", reset_components)()])
+            else:
+                reset_box = TimeBox.composite(
+                    [builder.get_implementation("reset", reset_components)()] * options.active_reset_cycles
+                )
+            new_boxes.append(reset_box + box)
+    except ValueError as exc:
+        raise ClientError(f"{exc}") from exc
     return new_boxes
 
 
@@ -670,10 +702,12 @@ def apply_dd_strategy(
         elif isinstance(options.dd_strategy, DDStrategy):
             strategy = options.dd_strategy
         else:
-            raise ValueError("Unsupported dynamical decoupling strategy submitted.")
+            raise UnknownCircuitExecutionOptionError("Unsupported dynamical decoupling strategy submitted.")
 
     else:
-        raise ValueError(f"Unsupported dynamical decoupling mode requested ({options.dd_mode}).")
+        raise UnknownCircuitExecutionOptionError(
+            f"Unsupported dynamical decoupling mode requested ({options.dd_mode})."
+        )
 
     new_schedules: list[Schedule] = []
 
@@ -729,7 +763,7 @@ def clean_schedule(schedules: Iterable[Schedule], builder: ScheduleBuilder) -> l
 @compiler_pass
 def build_playlist(schedules: Iterable[Schedule], builder: ScheduleBuilder) -> tuple[Playlist, dict[str, Any]]:
     """Build the playlist from the schedules."""
-    playlist = builder.build_playlist(schedules)
+    playlist = builder.build_playlist(schedules)[0]
     return playlist, {"schedules": schedules}  # save the schedules for building settings, debugging, etc
 
 
