@@ -1,4 +1,4 @@
-# Copyright 2024 IQM
+# Copyright 2024-2025 IQM
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,7 +15,7 @@
 """Utility functions for IQM Pulla."""
 
 from collections import namedtuple
-from collections.abc import Hashable, Iterable, Sequence
+from collections.abc import Hashable, Iterable, Sequence, Set
 from dataclasses import replace
 from itertools import chain
 from typing import Any
@@ -25,7 +25,7 @@ import numpy as np
 from exa.common.data.setting_node import SettingNode
 from exa.common.data.value import ObservationValue
 from exa.common.qcm_data.chip_topology import ChipTopology
-from iqm.cpc.compiler.errors import InsufficientContextError
+from iqm.cpc.compiler.errors import CalibrationError, InsufficientContextError, UnknownCircuitExecutionOptionError
 from iqm.cpc.compiler.station_settings import build_station_settings
 from iqm.cpc.interface.compiler import Circuit as CPC_Circuit
 from iqm.cpc.interface.compiler import (
@@ -42,7 +42,6 @@ from iqm.pulse.playlist.channel import ChannelProperties
 from iqm.pulse.playlist.instructions import Instruction
 from iqm.pulse.playlist.schedule import Schedule, Segment
 from iqm.pulse.timebox import TimeBox
-from iqm.station_control.client import utils as station_control_client_utils
 from iqm.station_control.interface.models.observation import ObservationBase
 
 LOCUS_SEPARATOR = "__"  # EXA uses this, currently
@@ -87,7 +86,7 @@ def iqm_circuit_to_gate_implementation(circuit: CPC_Circuit, qubit_mapping: dict
         # TODO This method does not heed GateImplementation.locus, and will always apply the
         # gate on fixed qubits. It just pastes ``circuit`` contents, with the qubits mapped using
         # the likewise fixed mapping, into a TimeBox.
-        registered_gates = list({instr.name for instr in circuit.instructions})
+        registered_gates = tuple({instr.name for instr in circuit.instructions})
 
         def __call__(self):
             boxes = []
@@ -417,12 +416,12 @@ def calset_to_cal_data_tree(calibration_set: CalibrationSet) -> OpCalibrationDat
         path = key.split(".")
         if path[0] == "gates":
             if len(path) < 5:
-                raise ValueError(f"Calibration observation name '{key}' is malformed.")
+                raise CalibrationError(f"Calibration observation name '{key}' is malformed.")
             # treat the locus specially
             locus = tuple(path[3].split(LOCUS_SEPARATOR))
             locus = () if locus == ("",) else locus
             # mypy likes this
-            set_path(tree.setdefault(path[1], {}).setdefault(path[2], {}).setdefault(locus, {}), path[4:], value)
+            set_path(tree.setdefault(path[1], {}).setdefault(path[2], {}).setdefault(locus, {}), path[4:], value)  # type: ignore[arg-type]
     return tree
 
 
@@ -457,7 +456,7 @@ def initialize_schedule_builder(
     return builder
 
 
-def _update_channel_props_from_calibration(
+def _update_channel_props_from_calibration(  # noqa: ANN202
     channel_properties: dict[str, ChannelProperties],
     component_channels: dict[str, dict[str, str]],
     calset: CalibrationSet,
@@ -480,11 +479,11 @@ def _update_channel_props_from_calibration(
             if center_frequency is None:
                 center_frequency = calset.get(f"controllers.{component}.readout.local_oscillator.frequency")
             if center_frequency is None:
-                raise ValueError(
+                raise CalibrationError(
                     f"No calibration value found for the center frequency or local oscillator frequency of {component}."
                 )
             channel_name = channels["readout"]
-            replacements[channel_name] = replace(channel_properties[channel_name], center_frequency=center_frequency)
+            replacements[channel_name] = replace(channel_properties[channel_name], center_frequency=center_frequency)  # type: ignore[call-arg]
 
     return channel_properties | replacements
 
@@ -494,7 +493,7 @@ def find_circuit_boundary(
     circuit_components: set[str] | frozenset[str],
     circuit_couplers: set[str],
     device: ChipTopology,
-) -> tuple[set[str], set[str]]:
+) -> tuple[Set[str], Set[str]]:
     """Determine the boundary of a circuit executed on the QPU.
 
     See :class:`.CircuitBoundaryMode` for the definitions of the circuit boundaries.
@@ -509,7 +508,7 @@ def find_circuit_boundary(
         boundary locus components, boundary couplers
 
     Raises:
-        ValueError: unknown ``mode``
+        UnknownCircuitExecutionOptionError: unknown ``mode``
 
     """
     if mode == CircuitBoundaryMode.NEIGHBOUR:
@@ -519,10 +518,10 @@ def find_circuit_boundary(
         }
     elif mode == CircuitBoundaryMode.ALL:
         # maybe safer/better: all unused locus components/couplers are considered boundary
-        boundary_components = (device.qubits | device.computational_resonators) - circuit_components
-        boundary_couplers = device.couplers - circuit_couplers
+        boundary_components = (device.qubits | device.computational_resonators) - circuit_components  # type: ignore[assignment]
+        boundary_couplers = device.couplers - circuit_couplers  # type: ignore[assignment]
     else:
-        raise ValueError(f"Unknown circuit boundary mode '{str(mode)}'")
+        raise UnknownCircuitExecutionOptionError(f"Unknown circuit boundary mode '{str(mode)}'")
     return boundary_components, boundary_couplers
 
 
@@ -602,4 +601,4 @@ def calset_from_observations(calset_observations: Iterable[ObservationBase]) ->
         calibration set
 
     """
-    return station_control_client_utils.calset_from_observations(calset_observations)
+    return {obs.dut_field: obs.value for obs in calset_observations}
